#include <stdlib.h>
#include <string.h>

static struct kmc_trigger {
	char *str;
	void (*cb)(void *, char *);
	void *cbdata;
} **triggers = NULL;
static int numtriggers = 0;

void (*kmc_register_hook)(char *, void (*)(void *, char *), void *) = NULL;
void (*kmc_unregister_hook)(char *) = NULL;
void (*kmc_query_hook)(char *, void (*)(int, void *), void *) = NULL;

static struct kmc_trigger *makekmct(char *str,
																		void (*cb)(void *, char *),
																		void *data)
{
	struct kmc_trigger *t = malloc(sizeof(*t));
	memset(t, 0, sizeof(*t));
	t->str = strdup(str);
	t->cb = cb;
	t->cbdata = data;
	return t;
}

void kmc_register_trigger(char *str, void (*cb)(void *, char *), void *data) {
	triggers = realloc(triggers, (++numtriggers) * sizeof(*triggers));
	triggers[numtriggers - 1] = makekmct(str, cb, data);
	if(kmc_register_hook)
		kmc_register_hook(str, cb, data);
}

void kmc_call_trigger(char *str, char *arg) {
	int i;

	for(i = 0; i < numtriggers; i++) {
		if(!strcmp(triggers[i]->str, str)) {
			triggers[i]->cb(triggers[i]->cbdata, arg);
		}
	}
}

void kmc_unregister_trigger(char *str) {
	int i;

	for(i = 0; i < numtriggers; i++)
		if(!strcmp(triggers[i]->str, str)) {
			free(triggers[i]->str);
			free(triggers[i]);
			triggers[i] = triggers[--numtriggers];
		}
	if(kmc_unregister_hook)
		kmc_unregister_hook(str);
}

void kmc_enumerate_triggers(void (*cb)(char *,
																			 void (*)(void *, char *),
																			 void *))
{
	int i;

	for(i = 0; i < numtriggers; i++)
		cb(triggers[i]->str, triggers[i]->cb, triggers[i]->cbdata);
}

void kmc_query(char *str, void (*cb)(int, void *), void *data) {
	if(kmc_query_hook)
		kmc_query_hook(str, cb, data);
	else
		cb(1, data);
}
