#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <sysexits.h>
#include <stdio.h>

#include <event.h>

struct callback {
  void (*func)(int, void *);
  void *data;
};

static fd_set rfds;
static int hirfd = -1;
static struct callback *rc;

void (*hookregr)(int, void (*)(int, void *), void *);
void (*hookunregr)(int);
void (*hookdispatch)();

void ev_enumr(void (*func)(int, void (*)(int, void *), void *)) {
	int i;

	for(i = 0; i <= hirfd; i++)
		if(FD_ISSET(i, &rfds))
			func(i, rc[i].func, rc[i].data);
}

static void ev_grow(int limit, int *old, struct callback **c) {
  if(limit <= *old)
    return;
  *old = limit;
  *c = realloc(*c, sizeof(**c) * (limit + 1));
}

static void ev_shrink(fd_set fds, int *old, struct callback **c) {
  int i;

  i = *old;
	if(FD_ISSET(i, &fds))
		return;
  while(i-- > 0)
    if(FD_ISSET(i, &fds)) {
      *old = i;
      *c = realloc(*c, sizeof(**c) * (*old + 1));
			return;
    }
}
      
void ev_unregr(int fd) {
  FD_CLR(fd, &rfds);
  ev_shrink(rfds, &hirfd, &rc);
	if(hookunregr)
		hookunregr(fd);
}

void ev_regr(int fd, void (*callback)(int, void *), void *data) {
  FD_SET(fd, &rfds);
  ev_grow(fd, &hirfd, &rc);
  rc[fd].func = callback;
  rc[fd].data = data;
	if(hookregr)
		hookregr(fd, callback, data);
}

void ev_init() {
  rc = NULL;
  hirfd = -1;
  FD_ZERO(&rfds);
}

void ev_dispatch() {
  fd_set fds;
  int i;

	while(1) {
		if(hookdispatch)
			hookdispatch();
		FD_ZERO(&fds);
		for(i = 0; i <= hirfd; i++)
			if(FD_ISSET(i, &rfds))
				FD_SET(i, &fds);
		if((i = select(hirfd + 1, &fds, NULL, NULL, NULL)) < 0)
			exit(EX_OSERR);
		for(i = 0; i <= hirfd; i++)
			if(FD_ISSET(i, &fds) && FD_ISSET(i, &rfds)) {
				rc[i].func(i, rc[i].data);
			}
	}
}
