#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <keymgr.h>
#include <event.h>

static char *sockpath = NULL;

int km_makefifo(char *str, char *env) {
	int fd;
	int pathlen = strlen(sockpath) + strlen(str) + 4;
	char *path = alloca(pathlen);

	snprintf(path, pathlen, "%s/%s", sockpath, str);
  unlink(path);
	if(mkfifo(path, S_IRWXU)) {
		perror("mkfifo");
		exit(EX_OSERR);
	}
	if((fd = open(path, O_RDONLY|O_NOCTTY|O_NONBLOCK)) < 0) {
		perror("open");
		exit(EX_OSERR);
	}
	setenv(env, path, 1);
	return fd;
}

int km_makesock(char *str, char *env) {
  struct sockaddr_un sun;
  int fd;

  if((fd = socket(PF_LOCAL, SOCK_STREAM, 0)) < 0) {
    perror("socket");
    exit(EX_OSERR);
  }
  fcntl(fd, F_SETFL, O_NDELAY);
  memset(&sun, 0, sizeof(sun));
  sun.sun_family = PF_LOCAL;
	/* how do we find out the size of sun_path? */
	sprintf(sun.sun_path, "%s/%s", sockpath, str);
  unlink(sun.sun_path);
  if(bind(fd, &sun, sizeof(sun))) {
    perror("bind");
    exit(EX_CANTCREAT);
  }
  if(listen(fd, 2)) {
    perror("listen");
    exit(EX_OSERR);
  }
	setenv(env, sun.sun_path, 1);
	return fd;
}

void reap() {
	exit(0);
}

int main(int argc, char **argv) {
	struct sigaction sa;
	int i;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = SIG_IGN;
	if(sigaction(SIGPIPE, &sa, NULL))
		perror("sigaction");
	sa.sa_handler = reap;
	if(sigaction(SIGCHLD, &sa, NULL))
		perror("sigaction");
	sockpath = malloc(80);
	sockpath[0] = 0;
	snprintf(sockpath, 79, "/tmp/.keymgr-%d", getpid());
	if(mkdir(sockpath, S_IRWXU)) {
		perror("mkdir");
		exit(EX_OSERR);
	}
	km_rc();
  switch(fork()) {
    case -1:
      perror("fork");
      exit(EX_OSERR);
    case 0:
			for(i = 3; i < getdtablesize(); i++)
				close(i);
      execvp(argv[1], argv + 1);
      perror("execvp");
      exit(EX_DATAERR);
    default:
      break;
  }
	close(0);
	close(1);
	close(2);
	setsid();
	ev_dispatch();
	return 0;
}
