#include <stdlib.h>

#include <console.h>

#include "librsa.h"
#include "queue.h"

static int numkeys = 0;
static struct rlq *keys = NULL;

void kmlr_addkey(struct kmlr_rsakey *key) {
  queue_add(&keys, key);
  numkeys++;
}

int kmlr_numkeys() { return numkeys; }

static int rsapub_eq(struct kmlr_rsapub *a, struct kmlr_rsapub *b) {
  return (a->len == b->len) && !mpz_cmp(a->e, b->e) && !mpz_cmp(a->n, b->n);
}
    
static struct kmlr_rsakey *findkey(struct kmlr_rsapub *key) {
  struct rlq *q = keys;

  while(q) {
    if(rsapub_eq(&((struct kmlr_rsakey *)q->data)->pub, key))
      return q->data;
    q = q->next;
  }
  return NULL;
}

struct rlq *kmlr_list() {
  struct rlq *ret = NULL;
  struct rlq *tmp = keys;

  while(tmp) {
    struct kmlr_rsakey *k = (struct kmlr_rsakey *)tmp->data;
    queue_add(&ret, &k->pub);
    tmp = tmp->next;
  }
  return ret;
}

struct querycb {
	struct kmlr_rsapub *pub;
	mpz_t ct;
	kmlr_dcb callback;
	void *cbdata;
};

static struct querycb *
makequerycb(struct kmlr_rsapub *pub, mpz_t ct, kmlr_dcb cb, void *data)
{
	struct querycb *ret = malloc(sizeof(*ret));

	ret->callback = cb;
	ret->cbdata = data;
	ret->pub = pub;
	mpz_init_set(ret->ct, ct);
	return ret;
}

static void kmlr_default_decrypt(int allow, void *data) {
	struct querycb *cb = data;
  struct kmlr_rsakey *key = findkey(cb->pub);
	mpz_t ret;

	mpz_init(ret);
  if(allow && key) {
		if(key->decrypt) {
			key->decrypt(cb->pub, cb->ct, cb->callback, cb->cbdata, key->decryptdata);
			return;
		}	else 
			mpz_powm(ret, cb->ct, key->priv.d, key->pub.n); /* DIY RSA */
	}
	cb->callback(cb->pub, ret, cb->cbdata);
	free(cb);
  return;
}

void kmlr_decrypt(struct kmlr_rsapub *pub, mpz_t ciphertext,
									kmlr_dcb callback, void *data)
{
  struct kmlr_rsakey *key = findkey(pub);
	void (*query)(char *, void (*)(int, void *), void *) = kmc_query;

	if(key->queryhook)
		query = key->queryhook;
	query(key->pub.comment, kmlr_default_decrypt,
				makequerycb(pub, ciphertext, callback, data));
}
