#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include <ds2480.h>
#include <crypto.h>
#include <string.h>

#include <console.h>
#include <string-num.h>

#include "../librsa.h"

struct decryptcb {
	struct ow_crypto *dev;
	int keynum;
};

static struct decryptcb *makedcb(struct ow_crypto *c, int num) {
	struct decryptcb *dcb = malloc(sizeof(*dcb));

	memset(dcb, 0, sizeof(*dcb));
	dcb->dev = c;
	dcb->keynum = num;
	return dcb;
}

/* straight out of ssh-pipe.c - need to move it somewhere common */

int kmsp_mpz_bin(mpz_t mp, char *buf, int *buflen) {
  int bytes, bp, hexbytes, hp, bits;
  char *hex;

  bits = mpz_sizeinbase(mp, 2);
  bytes = (bits + 7) / 8;
  if(*buflen < bytes)
    return -1;
  *buflen = bytes;
  hexbytes = mpz_sizeinbase(mp, 16);
  if(!(hex = alloca(hexbytes + 1)))
    return -1;
  mpz_get_str(hex, 16, mp);
  bp = 0;
  hp = 0;
  if(hexbytes & 1)
    buf[bp++] = kmsl_char_nybble(hex[hp++]);
  while(hp < hexbytes) {
    buf[bp++] = kmsl_str_byte(hex + hp);
    hp += 2;
  }
  return bits;
}

static void absorb(int ok, void *key) {
	if(ok) {
		kmlr_addkey(key);
	} else {
		/* wipe the key. should probably be in kmlr_... */
	}
}

static void *readarray(struct ow_crypto *c, int num, int *rlen, int temp) {
	struct CommandAPDU ca;
	struct ResponseAPDU ra;

	ca.cla = 0x80;
	ca.ins = temp ? 2 : 7;
	ca.p1 = 0;
	ca.p2 = num;
	ca.Lc = 0;
	ca.Le = 0;
	c->dispatchAPDU(c, &ca, &ra, 1);
	if(ra.sw != 0x9000) {
		*rlen = 0;
		return NULL;
	}
	*rlen = ra.Lr;
	return ra.data;
}

static void readnum(mpz_t ret, struct ow_crypto *c, int num, int temp) {
	int len, i;
	char *str;
	u_int8_t *data;

	data = readarray(c, num, &len, temp);
	str = alloca(2 * len + 1);
	for(i = 0; i < len; i++)
		memcpy(str + i * 2, kmsl_byte_str(data[i]), 2);
	str[2 * i] = 0;
	mpz_set_str(ret, str, 16);

}

static void decrypt(struct kmlr_rsapub *pub, mpz_t ct,
										kmlr_dcb cb, void *cbdata,
										void *d)
{
	struct CommandAPDU ca;
	struct ResponseAPDU ra;
	int len = 128;
	u_int8_t *data = alloca(len);
	struct decryptcb *dcb = d;
	mpz_t ret;

	mpz_init(ret);
	kmsp_mpz_bin(ct, data, &len);
	ca.cla = 0x80;
	ca.ins = 0;
	ca.p1 = dcb->keynum;
	ca.p2 = 5;
	ca.Lc = 64;
	ca.data = data;
	ca.Le = 0;
	dcb->dev->dispatchAPDU(dcb->dev, &ca, &ra, 1);
	if(ra.sw != 0x9000) {
		goto out;
	}
	ca.ins = 1;
	ca.data = data + 64;
	dcb->dev->dispatchAPDU(dcb->dev, &ca, &ra, 1);
	if(ra.sw != 0x9000) {
		goto out;
	}
	ca.ins = 4;
	ca.Lc = 0;
	dcb->dev->dispatchAPDU(dcb->dev, &ca, &ra, 7);
	if(ra.sw != 0x9000) {
		goto out;
	}
	readnum(ret, dcb->dev, 4, 1);
 out:
	cb(pub, ret, cbdata);
}

static void snarf(int ok, void *dev) {
	struct CommandAPDU ca;
	struct ResponseAPDU ra;
	struct ow_crypto *c = dev;
	u_int8_t aid[16] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16 };
	int i, num;
	mpz_t ret;
	struct kmlr_rsakey *key;
	char *qprefix = "Absorb key id: ";

	if(!ok)
		return;
	ca.cla = 0;
	ca.ins = 0xa4;
	ca.p1 = 4;
	ca.p2 = 0;
	ca.Lc = 16;
	ca.data = aid;
	ca.Le = 0;
	c->dispatchAPDU(c, &ca, &ra, 0);
	if(ra.sw != 0x9000) {
		return;
	}
	ca.cla = 0x80;
	ca.ins = 6;
	ca.p1 = 0;
	ca.p2 = 0;
	ca.Lc = 0;
	c->dispatchAPDU(c, &ca, &ra, 0);
	if((ra.sw != 0x9000) || (ra.Lr != 1)) {
		return;
	}
	num = ra.data[0];
	mpz_init(ret);
	for(i = 0; i < ra.data[0]; i++) {
		char *query, *id;
		int idlen, qlen;
		
		key = malloc(sizeof(*key));
		memset(key, 0, sizeof(*key));
		readnum(ret, c, 0, 0);
		mpz_init_set(key->pub.n, ret);
		key->pub.len = mpz_sizeinbase(ret, 2);
		readnum(ret, c, 2, 0);
		mpz_init_set(key->pub.e, ret);
		id = readarray(c, 6, &idlen, 0);
		key->pub.comment = malloc(idlen + 1);
		memcpy(key->pub.comment, id, idlen);
		key->pub.comment[idlen] = 0;
		key->decrypt = decrypt;
		key->decryptdata = makedcb(c, i);
		qlen = strlen(qprefix) + idlen + 1;
		query = alloca(qlen);
		strcpy(query, qprefix);
		strncat(query, id, idlen);
		query[qlen - 1] = 0;
		kmc_query(query, absorb, key);
	}
	mpz_clear(ret);
}
	
static void scan(void *port, char *str) {
	struct ow_port *p = port;
	struct ow_dev *d;

	d = p->scan(p);
	p->setspeed(p, SPD_OVERDRIVE);
	while(d) {
		if((d->serial[0] == 0x96) || (d->serial[0] == 0x16)) {
			char *line;
			struct ow_crypto *cib = d->private;

			line = alloca(200);
			snprintf(line, 199,
							 "Scan iButton %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x for RSA keys?",
							 d->serial[0], d->serial[1], d->serial[2], d->serial[3],
							 d->serial[4], d->serial[5], d->serial[6], d->serial[7]);
			kmc_query(line, snarf, cib);
		}
		d = d->next;
	}
}


void km_init(char *arg) {
	char *path = "/dev/ttyS0";
	struct ow_port *p;

	if(arg)
		path = arg;
	if((p = ds2480_make(path)))
		kmc_register_trigger("Scan-JiBs", scan, p);
}
