#include <gmp.h>
#include <stdlib.h>
#include <alloca.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <string.h>

#include <string-num.h>
#include <ssh-pipe.h>

int kmsp_mpz_bin(mpz_t mp, char *buf, int *buflen) {
  int bytes, bp, hexbytes, hp, bits;
  char *hex;

  bits = mpz_sizeinbase(mp, 2);
  bytes = (bits + 7) / 8;
  if(*buflen < bytes)
    return -1;
  *buflen = bytes;
  hexbytes = mpz_sizeinbase(mp, 16);
  if(!(hex = alloca(hexbytes + 1)))
    return -1;
  mpz_get_str(hex, 16, mp);
  bp = 0;
  hp = 0;
  if(hexbytes & 1)
    buf[bp++] = kmsl_char_nybble(hex[hp++]);
  while(hp < hexbytes) {
    buf[bp++] = kmsl_str_byte(hex + hp);
    hp += 2;
  }
  return bits;
}

int kmsp_mpz_bin32(mpz_t mp, char *buf, int *buflen) {
  int bits, bytes, offset = 0, newlen;
  
  bits = mpz_sizeinbase(mp, 2);
  bytes = (bits + 7) / 8;
  if(*buflen < bytes)
    return -1;
  if(bytes & 0x3)
    offset = 4 - (bytes & 0x3);
  newlen = *buflen - offset;
  kmsp_mpz_bin(mp, buf + offset, &newlen);
  *buflen = newlen + offset;
  while(offset--)
    buf[offset] = '\0';
  return bits;
}

int kmsp_buf_u32(u_int32_t *out, char *buf) {
  *out = ntohl(*((u_int32_t *)buf));
  return 4;
}

int kmsp_buf_u16(u_int16_t *out, char *buf) {
  *out = ntohs(*((u_int16_t *)buf));
  return 2;
}

int kmsp_buf_mpz(mpz_t *out, char *buf) {
  int offset = 0, i, bytes;
  u_int16_t bits;
  char *hexbuf, *tmp;

  offset += kmsp_buf_u16(&bits, buf + offset);
  bytes = (bits + 7) / 8;
   hexbuf = alloca(2 * bytes + 1);
  for(i = 0; i < bytes; i++) {
    tmp = kmsl_byte_str(buf[offset + i]);
    hexbuf[2 * i] = tmp[0];
    hexbuf[2 * i + 1] = tmp[1];
  }
  hexbuf[2 * i] = '\0';
  offset += bytes;
  mpz_set_str(*out, hexbuf, 16);
  return offset;
}

int kmsp_buf_buf(char *out, char *in, int len) {
  memcpy(out, in, len);
  return len;
}

int kmsp_buf_str(char *out, char *in, int *len) {
  int offset = 0;
  u_int32_t str;

  offset += kmsp_buf_u32(&str, in + offset);
  if((*len + 1) < str)
    abort();
  offset += kmsp_buf_buf(out, in + offset, str);
  out[str] = '\0';
  *len = str + 1;
  return offset;
}

int kmsp_buf_print(char *out, char *in, int len) {
  int i;
  char *tmp;

  for(i = 0; i < len; i++) {
    tmp = kmsl_byte_str(in[i]);
    out[2 * i] = tmp[0];
    out[2 * i + 1] = tmp[1];
  }
  return len;
}
