#include <unistd.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <gmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <stdio.h>
#define _GNU_SOURCE
#include <string.h>

#include <keymgr.h>

#include "ssh-authfd.h"
#include "ssh-pipe.h"
#include "md5.h"
#include "librsa.h"
#include "queue.h"
#include "event.h"

static unsigned int cookie = 1;
static unsigned int *fdcookie = NULL;
static int hifdcookie = -1;
static int dump = 0;
static void dump_mpz(char *str, mpz_t n);

struct netbuf {
  char *buf;
  int head;
  int size;
};

struct decryptcb {
	char sid[16];
	int fd;
	int cookie;
};

static struct decryptcb *makecb(char *sid, int fd) {
	struct decryptcb *d = malloc(sizeof(*d));

	memcpy(d->sid, sid, 16);
	d->fd = fd;
	d->cookie = fdcookie[fd];
	return d;
}

void ssh_decrypt(struct kmlr_rsapub *key, mpz_t pt, void *d) {
	int tmp, j, majiclen;
	char *challenge, *majic2, *ptr;
	struct MD5Context md;
	char out[16];
	struct iovec outv[3];
	u_int32_t msglen;
	u_int8_t type = SSH_AGENT_FAILURE;
	struct decryptcb *data = d;
	
	if(dump)
		dump_mpz("response: ", pt);
	if(data->cookie != fdcookie[data->fd])
		goto out;
	tmp = (key->len + 7) / 8;
	challenge = alloca(tmp);
	kmsp_mpz_bin32(pt, challenge, &tmp);

	if(challenge[0] != 0 || challenge[1] != 2)
		goto outwrite;

				/* lose everything up to and including the first null after magic */
	if(!(ptr = memchr(challenge + 2, 0, tmp - 2)))
		goto outwrite;
	j = ptr - (challenge + 2);
	mpz_mod_2exp(pt, pt, 8 * (tmp - j - 3));
	
	majiclen = 32;
	majic2 = alloca(32);
	if(kmsp_mpz_bin(pt, majic2, &majiclen) < 0)
		goto outwrite;

	MD5Init(&md);
	MD5Update(&md, majic2, 32);
	MD5Update(&md, data->sid, 16);
	MD5Final(out, &md);

	type = SSH_AGENT_RSA_RESPONSE;
 outwrite:
	outv[0].iov_base = &msglen;
	outv[0].iov_len = sizeof(msglen);
	outv[1].iov_base = &type;
	outv[1].iov_len = sizeof(type);
	outv[2].iov_base = out;
	outv[2].iov_len = 16;
	msglen = htonl(1 + 16);
	writev(data->fd, outv, 3);
 out:
	mpz_clear(pt);
	mpz_clear(key->n);
	mpz_clear(key->e);
	free(key);
	free(d);
}

int zenlike_acceptance(int fd, struct netbuf *d) {
  u_int32_t len, buflen;
  char type;
  int bits;
  char *buf = d->buf;
  int offset;

  buflen = d->head;
  if(buflen < 5)
    return 1;
  offset = 0;
  len = ntohl(len);
  offset += kmsp_buf_u32(&len, buf + offset);
  type = buf[offset++];
  switch(type) {
    case SSH_AGENTC_RSA_CHALLENGE:
      {
				u_int32_t resp;
				mpz_t ct;
				char sid[16];
				struct kmlr_rsapub *key = malloc(sizeof(*key));

				memset(key, 0, sizeof(*key));
				if(buflen == 5)
					return 1;
				offset += kmsp_buf_u32(&key->len, buf + offset);
				mpz_init(key->e);
				offset += kmsp_buf_mpz(&key->e, buf + offset);
				mpz_init(key->n);
				offset += kmsp_buf_mpz(&key->n, buf + offset);
				mpz_init(ct);
				offset += kmsp_buf_mpz(&ct, buf + offset);
				offset += kmsp_buf_buf(sid, buf + offset, 16);
				offset += kmsp_buf_u32(&resp, buf + offset);
				/* cooking with gas */
				if(dump)
					dump_mpz("challenge: ", ct);
				kmlr_decrypt(key, ct, ssh_decrypt, makecb(sid, fd));
				mpz_clear(ct);
			}
      break;
    case SSH_AGENTC_REQUEST_RSA_IDENTITIES:
      {
				struct iovec out[3];
				u_int32_t tml, msglen = 0, numid;
				u_int8_t type = SSH_AGENT_RSA_IDENTITIES_ANSWER;
				u_int16_t elen, nlen;
				int i, num;
				struct rlq *q, *outq = NULL;

				out[0].iov_base = &tml;
				out[0].iov_len = 4;
				out[1].iov_base = &type;
				out[1].iov_len = 1;
				msglen += 1;
				out[2].iov_base = &numid;
				out[2].iov_len = 4;
				msglen += 4;

				q = kmlr_list();
				num = 0;
				while(q) {
					struct iovec *outid = malloc(7 * sizeof(*outid));
					struct kmlr_rsapub *key = q->data;
					struct rlq *next = q->next;

					outid[0].iov_base = malloc(4);
					outid[0].iov_len = 4;
					msglen += 4;
					*(u_int32_t *)outid[0].iov_base = htonl(key->len);

					outid[1].iov_base = &elen;
					outid[1].iov_len = 2;
					msglen += 2;

					i = 8192;
					outid[2].iov_base = malloc(i);
					if((bits = kmsp_mpz_bin(key->e, outid[2].iov_base, &i)) < 0)
						exit(1);
					elen = htons(bits);
					outid[2].iov_len = i;
					msglen += i;

					outid[3].iov_base = &nlen;
					outid[3].iov_len = 2;
					msglen += 2;

					i = 8192;
					outid[4].iov_base = malloc(i);
					if((bits = kmsp_mpz_bin(key->n, outid[4].iov_base, &i)) < 0)
						exit(1);
					nlen = htons(bits);
					outid[4].iov_len = i;
					msglen += i;

					outid[5].iov_base = malloc(4);
					outid[5].iov_len = 4;
					msglen += 4;
					*(u_int32_t *)outid[5].iov_base = htonl(strlen(key->comment));

					outid[6].iov_base = key->comment;
					outid[6].iov_len = strlen(key->comment);
					msglen += strlen(key->comment);

					queue_add(&outq, outid);
					num++;
					free(q);
					q = next;
				}
				numid = htonl(num);
				tml = htonl(msglen);
				writev(fd, out, 3);
				while(outq) {
					struct rlq *next = outq->next;
					struct iovec *outid = (struct iovec *)outq->data;
					writev(fd, outid, 7);
					free(outid[0].iov_base);
					free(outid[2].iov_base);
					free(outid[4].iov_base);
					free(outid[5].iov_base);
					free(outid);
					free(outq);
					outq = next;
				}
      }
      break;
    case SSH_AGENTC_ADD_RSA_IDENTITY:
      {
				int comlen = 1024;
				u_int32_t bits;
				struct kmlr_rsakey *key = malloc(sizeof(*key));

				memset(key, 0, sizeof(*key));
				offset += kmsp_buf_u32(&bits, buf + offset);
				key->pub.len = bits;
				mpz_init(key->pub.n);
				offset += kmsp_buf_mpz(&key->pub.n, buf + offset);
				mpz_init(key->pub.e);
				offset += kmsp_buf_mpz(&key->pub.e, buf + offset);
				mpz_init(key->priv.d);
				offset += kmsp_buf_mpz(&key->priv.d, buf + offset);
				mpz_init(key->priv.u);
				offset += kmsp_buf_mpz(&key->priv.u, buf + offset);
				mpz_init(key->priv.p);
				offset += kmsp_buf_mpz(&key->priv.p, buf + offset);
				mpz_init(key->priv.q);
				offset += kmsp_buf_mpz(&key->priv.q, buf + offset);
				key->pub.comment = malloc(comlen);
				offset += kmsp_buf_str(key->pub.comment, buf + offset, &comlen);
				kmlr_addkey(key);
				if(dump) {
					dump_mpz("n: ", key->pub.n);
					dump_mpz("e: ", key->pub.e);
					dump_mpz("d: ", key->priv.d);
				}
      }
      {
				struct iovec out[2];
				u_int32_t msglen;
				u_int8_t type;

				out[0].iov_base = &msglen;
				out[0].iov_len = sizeof(msglen);
				out[1].iov_base = &type;
				out[1].iov_len = sizeof(type);
				msglen = htonl(1);
				type = SSH_AGENT_SUCCESS;
				writev(fd, out, 2);
      }
      break;
    default:
      break;
  }
  return 0;
}


void ssh_read(int fd, void *d) {
  int i;
  struct netbuf *data = d;
  
  ev_unregr(fd);
  if(!data) {
    data = malloc(sizeof(*data));
    memset(data, 0, sizeof(*data));
    data->size = 8192;
    data->buf = malloc(data->size);
  }
  i = read(fd, data->buf + data->head, data->size - data->head);
  if(i <= 0) {
    free(data->buf);
    free(data);
    close(fd);
    return;
  }
  data->head += i;
  if(!zenlike_acceptance(fd, data)) {
    data->head = 0;
  }
  ev_regr(fd, ssh_read, data);
}

void ssh_accept(int fd, void *data) {
  int i;

  if((i = accept(fd, NULL, NULL)) < 0)
    return;
	if(i > hifdcookie) {
		hifdcookie = i;
		fdcookie = realloc(fdcookie, sizeof(*fdcookie) * (hifdcookie + 1));
	}
	fdcookie[i] = cookie++;
  ev_regr(i, ssh_read, NULL);
}

void km_init(char *arg) {
  int fd;
	char *pid = alloca(9);

	fd = km_makesock("ssh-agent", "SSH_AUTH_SOCK");
  ev_regr(fd, ssh_accept, NULL);
	snprintf(pid, 8, "%d", getpid());
	setenv("SSH_AGENT_PID", pid, 1);
	if(arg && !strncmp(arg, "dump", 4))
		dump = 1;
}

static void dump_mpz(char *str, mpz_t mp) {
	int hexbytes, i = 0;
	char *hex;
					
	hexbytes = mpz_sizeinbase(mp, 16);
	if(!(hex = alloca(hexbytes + 1)))
		return;
	mpz_get_str(hex, 16, mp);
	fprintf(stderr, "%s ", str);
	if(hexbytes & 1) {
		fprintf(stderr, "0x0%c ", hex[0]);
		i = 1;
	}
	while(i < hexbytes) {
		fprintf(stderr, "0x%c%c ", hex[i], hex[i + 1]);
		i += 2;
	}
	fputs("\n", stderr);
}

